from operator import itemgetter

from jinja2 import BaseLoader, FileSystemLoader, PackageLoader, TemplateNotFound


class MultiLoader(BaseLoader):
    """
    Similar to Jinja's ChoiceLoader with 2 notable differences:
      - Loaders can be added at runtime
      - Each loader has a priority so the caller can decide
        in which order the loaders are evaluated
    """

    def __init__(self) -> None:
        super().__init__()
        self._loaders: list[tuple[BaseLoader, int]] = []
        # add default loaders
        self.add_loader(PackageLoader("banks", "templates"), 10)
        self.add_loader(FileSystemLoader("templates"), 20)

    def add_loader(self, loader: BaseLoader, priority: int = 100) -> None:
        self._loaders.append((loader, priority))

    def get_source(self, environment, template):
        # Sort by priority, ascending
        for loader, _ in sorted(self._loaders, key=itemgetter(1)):
            try:
                return loader.get_source(environment, template)
            except TemplateNotFound:
                continue

        raise TemplateNotFound(template)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from unittest import mock

from banks import Prompt, env


def test_run_prompt():
    p = Prompt.from_template("run_prompt.jinja")
    env.extensions["banks.extensions.generate.GenerateExtension"]._generate = mock.MagicMock(return_value="foo")

    assert p.text({"topic": "climate change"}) == "\n\nfoo\n"


def test_run_prompt_process():
    p = Prompt.from_template("run_prompt_process.jinja")
    env.extensions["banks.extensions.generate.GenerateExtension"]._generate = mock.MagicMock(return_value="foo bar baz")

    assert p.text({"topic": "climate change"}) == "FOO BAR BAZ"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import os
from pathlib import Path
from unittest import mock

import pytest

from banks import Prompt, env


def _get_data(name):
    here = Path(os.path.dirname(os.path.abspath(__file__)))
    with open(here / "test_default_templates" / name) as f:
        return f.read()


def test_blog():
    p = Prompt.from_template("blog.jinja")
    assert _get_data("blog.jinja.out") == p.text({"topic": "climate change"})


def test_summarize():
    p = Prompt.from_template("summarize.jinja")
    documents = [
        "A first paragraph talking about AI",
        "A second paragraph talking about climate change",
        "A third paragraph talking about retrogaming",
    ]
    assert _get_data("summarize.jinja.out") == p.text({"documents": documents})


def test_summarize_lemma():
    pytest.importorskip("simplemma")

    p = Prompt.from_template("summarize_lemma.jinja")
    assert _get_data("summarize_lemma.jinja.out") == p.text({"document": "The cats are running"})


def test_generate_tweet():
    p = Prompt.from_template("generate_tweet.jinja")
    env.extensions["banks.extensions.generate.GenerateExtension"]._generate = mock.MagicMock(return_value="foo")

    assert _get_data("generate_tweet.jinja.out") == p.text({"topic": "climate change"})



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import cast

from jinja2 import nodes
from jinja2.ext import Extension
from litellm import ModelResponse, completion

DEFAULT_MODEL = "gpt-3.5-turbo"


class GenerateExtension(Extension):
    """
    `generate` can be used to call the OpenAI API passing the tag text as a prompt and get back some content.

    Example:
        ```
        {% generate "write a tweet with positive sentiment" "gpt-3.5-turbo" %}
        Feeling grateful for all the opportunities that come my way! #positivity #productivity
        ```
    """

    # a set of names that trigger the extension.
    tags = {"generate"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token so that we can give
        # that line number to the nodes we create by hand.
        lineno = next(parser.stream).lineno

        # The args passed to the extension:
        # - the prompt text used to generate new text
        # - (optional) the name of the model use to generate new text
        args = [parser.parse_expression()]

        return nodes.Output([self.call_method("_generate", args)]).set_lineno(lineno)

    def _generate(self, text, model_name=DEFAULT_MODEL):
        """
        Helper callback.

        To tweak the prompt used to generate content, change the variable `messages` .
        """
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": text},
        ]

        response: ModelResponse = cast(ModelResponse, completion(model=model_name, messages=messages))
        return response["choices"][0]["message"]["content"]



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import Optional

from banks.env import env


class Prompt:
    def __init__(self, text: str) -> None:
        self._template = env.from_string(text)

    @classmethod
    def from_template(cls, name: str) -> "Prompt":
        p = cls("")
        p._template = env.get_template(name)
        return p

    def text(self, data: Optional[dict] = None) -> str:
        data = data or {}
        return self._template.render(data)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from jinja2 import Environment, select_autoescape

from banks.extensions import GenerateExtension, HFInferenceEndpointsExtension
from banks.filters import lemmatize
from banks.loader import MultiLoader

# Init the Jinja env
env = Environment(
    loader=MultiLoader(),
    extensions=[GenerateExtension, HFInferenceEndpointsExtension],
    autoescape=select_autoescape(
        enabled_extensions=("html", "xml"),
        default_for_string=True,
    ),
    trim_blocks=True,
    lstrip_blocks=True,
)

# Setup custom filters
env.filters["lemmatize"] = lemmatize


def with_env(cls):
    """
    A decorator that adds an `env` attribute to the decorated class
    """
    cls.env = env
    return cls
